/*
 * Speed-up library for CluDisc local adjusting.
 * (C) Martin Saturka http://www.bioplexity.org/
 * Opensource, available under the MIT/X license.
 */

#include <math.h>

#if defined __USE_MISC || defined __USE_XOPEN_EXTENDED || defined __USE_ISOC99
#define WITHLOG1P 1
#else
#define WITHLOG1P 0
double log1p (double x)
{
  return log(1.0 + x);
}
#endif

void locadj(int ecnt, int extends, double dzero, double *distdiffs, double *adjusted,
            double *curmovex, double *curmovey, double *curmovez, double curaten)
{
  int ione, isec;
  double oridist, curdist, curmoved, currat;
  double xdiff, ydiff, zdiff, tdiff;
  double curtomove;

  double *pdistdiffs;
  double *padjusted1;
  double *padjusted2;

  double (*intfunc) (double);

  if (1 == extends) {
    intfunc = log1p;
  }
  else {
    intfunc = atan;
  }

  pdistdiffs = distdiffs;
  padjusted1 = adjusted;
  for (ione = 0; ione < (ecnt - 1); ione++) {
    padjusted2 = padjusted1 + 3;
    for (isec = (ione + 1); isec < ecnt; isec++) {
      oridist = pdistdiffs[isec];
      xdiff = padjusted1[0] - padjusted2[0];
      ydiff = padjusted1[1] - padjusted2[1];
      zdiff = padjusted1[2] - padjusted2[2];

      curdist = (xdiff * xdiff) + (ydiff * ydiff) + (zdiff * zdiff);
      curdist = sqrt(curdist);
      curmoved = 0.0;
      if (curdist < oridist) {
        curmoved = intfunc(oridist / dzero) - intfunc(curdist / dzero);
        curmoved *= dzero;
      }
      if (curdist > oridist) {
        currat = oridist / dzero;
        if (2 == extends) {
          currat *= currat;
        }
        curmoved = (oridist - curdist) / (1.0 + currat);
      }

      tdiff = fabs(xdiff) + fabs(ydiff) + fabs(zdiff);
      if (0.0 == tdiff) {
        xdiff = 1.0;
        ydiff = 1.0;
        zdiff = 1.0;
        tdiff = 3.0;
      }
      curtomove = (xdiff / tdiff) * curmoved;
      curmovex[ione] += curtomove;
      curmovex[isec] -= curtomove;
      curtomove = (ydiff / tdiff) * curmoved;
      curmovey[ione] += curtomove;
      curmovey[isec] -= curtomove;
      curtomove = (zdiff / tdiff) * curmoved;
      curmovez[ione] += curtomove;
      curmovez[isec] -= curtomove;

      padjusted2 += 3;
    }
    pdistdiffs += ecnt;
    padjusted1 += 3;
  }

  padjusted1 = adjusted;
  for (ione = 0; ione < ecnt; ione++) {
    padjusted1[0] += curmovex[ione] * curaten;
    padjusted1[1] += curmovey[ione] * curaten;
    padjusted1[2] += curmovez[ione] * curaten;

    padjusted1 += 3;
  }

  return;
}

