/*
 * Speed-up library for CluDisc clustered scaling.
 * (C) Martin Saturka http://www.bioplexity.org/
 * Opensource, available under the MIT/X license.
 */

#include <math.h>

int sdist(int mcols, double *center, double *onerow, double *resvalue);

int censearch(int mrows, int cencount, int mcols, double *pdata, double *centers,
              long *cenlive, long *cranks, long *orients)
{
  int changed = 0;
  int irow, icen, ccen, ioff;
  int orient;
  int runcen;

  double cdist;
  double newldist;

  double *actual = pdata;

  for (irow = 0; irow < mrows; irow++, actual += mcols) {
    cdist = 0.0;
    newldist = -1.0;
    orient = 0;

    ccen = cranks[irow];

    if (-1 != ccen) {
      ioff = ccen * mcols;
      orient = sdist(mcols, centers+ioff, actual, &cdist);
      orients[irow] = orient;
      newldist = cdist;
      if (!orient) {
        ccen = -1;
        cranks[irow] = -1;
        changed = 1;
      }
    }

    runcen = ccen;

    for (icen = 0, ioff=0; icen < cencount; icen++, ioff += mcols) {
      if (!(cenlive[icen])) {
        continue;
      }
      if (ccen == icen) {
        continue;
      }

      orient = sdist(mcols, centers+ioff, actual, &cdist);

      if (0 == orient) {
        continue;
      }

      if ((-1 == runcen) || (cdist < newldist)) {
        runcen = icen;
        newldist = cdist;
        cranks[irow] = icen;
        orients[irow] = orient;
        changed = 1;
      }
    }

  }

  return changed;

}

int sdist(int mcols, double *center, double *onerow, double *resvalue)
{
  int common = 0;
  double cpdist = 0.0;
  double cndist = 0.0;

  double vrow;
  double vcen;

  int icol;

  for (icol = 0; icol < mcols; icol++) {
    vrow = onerow[icol];
    vcen = center[icol];

    if ((vrow != vrow) || (vcen != vcen)) {
      continue;
    }

    common += 1;
    cpdist += (vrow - vcen) * (vrow - vcen);
    cndist += (vrow + vcen) * (vrow + vcen);
  }

  if (0 == common) {
    return 0;
  }

  if (cpdist <= cndist) {
    resvalue[0] = sqrt(cpdist / common);
    return 1;
  }
  else {
    resvalue[0] = sqrt(cndist / common);
    return -1;
  }

  return 0;
}

